# Nodejs的内存控制

### JavaScript中无法立即回收内存的两种情况：

1. 全局变量的主动释放

* 虽然delete操作和重新赋值具有相同的效果，但是在V8中通过delete删除对象的属性有可能干扰V8的优化，所以赋值方式解除引用更好。

```
global.foo = undefined; // or null
```

2. 闭包

* 实现外部作用域访问内部作用域中变量的方法叫做闭包（closure）。

* 一旦外部变量引用内部中间函数，则中间函数将不会释放,作用域中的变量不会释放（函数具备了访问具备作用域中的变量的问题），从而内存占用不会得到释放。

### 内存指标

1. 查看进程的内存占用
```bash
    $ node
    > process.memoryUsage()
    { rss: 13852672,//resident set size的缩写，即进程的常驻内存部分。
    heapTotal: 6131200,
    heapUsed: 2757120 }
```
2. 查看系统的内存占用
```bash
$ node
> os.totalmem()
8589934592
> os.freemem()
4527833088
>
```
3. 堆外内存

Buffer对象不经过V8内存分配机制，所以不会有堆内存的大小限制

### 内存泄漏

* 缓存意味着它将常驻在老生代中，缓存中存储的键越多，长期存活的对象也就越多，这将导致垃圾回收再进行扫描和整理时，堆这些对象做无用功。

```javascript
_.memoize = function(func, hasher) 
{
    var memo = {};
    hasher || (hasher = _.identity);
    return function() 
    {
    var key = hasher.apply(this, arguments);
    return _.has(memo, key) ? memo[key] : (memo[key] = func.apply(this, arguments));
    };
};
它的原理是以参数作为键进行缓存，以内存空间换CPU执行时间。这里潜藏的陷阱即是每个被执行的结果都会按参数缓存在memo对象上，不会被清除。
```
### 内存泄漏排查

1. node-heapdump

(1) 安装node-heapdump

  npm install heapdump

（2） 编写sever.js

```javascript
var heapdump = require('heapdump');
var leakArray = [];
var leak = function () {
leakArray.push("leak" + Math.random());
};
http.createServer(function (req, res) {
leak();
res.writeHead(200, {'Content-Type': 'text/plain'});
res.end('Hello World\n');
}).listen(1337);
console.log('Server running at http://127.0.0.1:1337/');

(3) 执行node server.js,打开网页访问多次

kill -USR2 <pid> //向服务进程发送Sigusr2信号。
抓取的文件将在heapdump-<sec>.<usec>.heapsnapsho的格式存放，这是一份较大的JSON文件，需要通过Chrome的开发者工具打开查看。
```
2. node-memwatch

（1） 先安装

（2）准备一段内存泄漏的代码：
```javascript
var memwatch = require('memwatch');
memwatch.on('leak', function (info) {
    console.log('leak:');
    console.log(info);
});
memwatch.on('stats', function (stats) {
    console.log('stats:')
    console.log(stats);
});
var http = require('http');
var leakArray = [];
var leak = function () {
   leakArray.push("leak" + Math.random());
};
http.createServer(function (req, res) {
    leak();
    res.writeHead(200, {'Content-Type': 'text/plain'});
    res.end('Hello World\n');
}).listen(1337);
console.log('Server running at http://127.0.0.1:1337/');  
```
（3）每次进行全堆垃圾回收时，将会触发一次states事件

（4） 连续5次垃圾回收后，内存依然没有被释放,这意味着有内存泄漏的产生，从而会触发一次leak事件。

（5）如果需要堆内存比较，修改成以下代码：见P152页
```javascript
var memwatch = require('memwatch');
var leakArray = [];
var leak = function () {
    leakArray.push("leak" + Math.random());
};
// Take first snapshot
var hd = new memwatch.HeapDiff();
for (var i = 0; i < 10000; i++) {
    leak();
}
// Take the second snapshot and compute the diff
var diff = hd.end();
console.log(JSON.stringify(diff, null, 2));
```

### 大内存应用

stream模块是Node的原生模块，可用于处理大文件，stream继承自EventEmitter。

（1） fs的createReadSream()和createWriteStream()
（2）process模块的的stdin􀖖stdout
```javascript
var reader = fs.createReadStream('in.txt');
var writer = fs.createWriteStream('out.txt');
reader.on('data', function (chunk) {
writer.write(chunk);
});
reader.on('end', function () {
    writer.end();
});
或者如下：
var reader = fs.createReadStream('in.txt');
var writer = fs.createWriteStream('out.txt');
reader.pipe(writer);
```

> 由于V8的内存限制，我们无法通过fs.readFile()和fs.writeFile()直接进行大文件的操作。